import numpy as np


def pps(knn, rnn, order):
    """
    Plum Pudding Sampling (PPS)
    This function returns the ID of landmarks.

    Parameters are:

    'knn'      - A N by k matrix. Each row represents the KNN of each point.
    'rnn'      - A vector of length N. Each row represents the RNN of each point.
    'order'    - A positive integer specifying the order of KNN. Once a landmark is selected, its KNN will be removed
                 from the queue. PPS supports the removal of multi-order KNN.

    """
    id_samp = []
    id_sort = sorted(range(len(rnn)), key=lambda k: rnn[k], reverse=True)
    while len(id_sort) > 0:
        id_samp.append(id_sort[0])
        rm_pts = [id_sort[0]]
        for _ in range(order):
            rm_pts.extend(knn[rm_pts].flatten().tolist())
        rm_pts = set(rm_pts)
        rm_id = np.where(np.isin(id_sort, list(rm_pts)))[0]
        id_sort = [id_sort[i] for i in range(len(id_sort)) if i not in rm_id]
    return id_samp



from sklearn.preprocessing import MinMaxScaler
from sklearn.neighbors import NearestNeighbors
from .init_pca import init_pca
from .pps import pps
from .learning_s import learning_s
from .learning_l import learning_l
from .opt_scale import opt_scale
from .clle import clle
import numpy as np


def scml(
        X,
        no_dims=2,
        k1=20,
        normalize=True,
        large=False,
        initialize='le',
        agg_coef=1.2,
        T_epoch=50
):
    """
    This function returns representation of the N by D matrix X in the lower-dimensional space. Each row in X
    represents an observation.

    Parameters are:

    'no_dims'      - A positive integer specifying the number of dimension of the representation Y.
                   Default: 2
    'k1'           - A non-negative integer specifying the number of nearest neighbors for PPS to
                   sample landmarks. It must be smaller than N.
                   Default: adaptive
    'normalize'    - Logical scalar. If true, normalize X using min-max normalization. If features in
                   X are on different scales, 'Normalize' should be set to true because the learning
                   process is based on nearest neighbors and features with large scales can override
                   the contribution of features with small scales.
                   Default: True
    'large'        - Logical scalar. If true, the data can be split into multiple blocks to avoid the problem
                   of memory overflow, and the gradient can be computed block by block using 'learning_l' function.
                   Default: False
    'initialize'   - A string specifying the method for initializing Y before manifold learning.
        'le'       - Laplacian eigenmaps.
        'pca'      - Principal component analysis.
        'mds'      - Multidimensional scaling.
                   Default: 'le'
    'agg_coef'     - A positive scalar specifying the aggregation coefficient.
                   Default: 1.2
    'T_epoch'      - Maximum number of epochs to take.
                   Default: 50

    """
    # Remove duplicate observations
    X, orig_id = np.unique(X, axis=0, return_inverse=True)

    # Obtain size and dimension of data
    n, dim = X.shape

    # Normalize the data
    if normalize:
        X = MinMaxScaler().fit(X).transform(X)

    # Perform PPS to obtain the landmarks
    if k1 > 0:
        if n > 5000 and dim > 50:
            xx = init_pca(X, no_dims, 0.8)
            get_knn = NearestNeighbors(n_neighbors=k1 + 1).fit(xx).kneighbors(xx, return_distance=False)
        else:
            get_knn = NearestNeighbors(n_neighbors=k1 + 1).fit(X).kneighbors(X, return_distance=False)
        _, rnn = np.unique(get_knn, return_counts=True)
        id_samp = pps(get_knn, rnn, 1)
    else:
        get_knn = []
        rnn = []
        id_samp = list(range(n))

    X_samp = X[id_samp]

    # Compute embedding of landmarks
    if not large:
        Y_samp, k2 = learning_s(X_samp, k1, get_knn, rnn, id_samp, no_dims, initialize, agg_coef, T_epoch)
    else:
        Y_samp, k2 = learning_l(X_samp, k1, get_knn, rnn, id_samp, no_dims, initialize, agg_coef, T_epoch)

    # Compute embedding of non-landmarks
    if k1 > 0:
        id_rest = np.setdiff1d(range(n), id_samp)
        X_rest = X[id_rest]
        Y_rest = np.zeros((len(id_rest), no_dims))
        # Compute the optimal scale for each landmark
        scale = opt_scale(X_samp, Y_samp, k2)
        top_k = no_dims + 1
        near_dis, near_samp = NearestNeighbors(n_neighbors=top_k).fit(X_samp).kneighbors(X_rest)
        for i in range(len(id_rest)):
            near_top_k = near_samp[i]
            top_X = X_samp[near_top_k]
            top_Y = Y_samp[near_top_k]
            N_dis = near_dis[i, 0] * scale[near_top_k[0]]
            # Perform CLLE
            Y_rest[i] = clle(top_X, top_Y, X_rest[i], N_dis)
        YY = np.zeros((n, no_dims))
        YY[id_rest] = Y_rest
        YY[id_samp] = Y_samp
    else:
        YY = Y_samp
    Y = YY[orig_id]

    return Y



import numpy as np


def pca(X, no_dims):
    """
     This function performs PCA embedding.

     Parameters are:

     'X'          - N by D matrix. Each row in X represents an observation.
     'no_dims'    - A positive integer specifying the number of dimension of the representation Y.

     """
    X = X - np.mean(X, axis=0)
    C = np.cov(X, rowvar=False)
    C[np.isnan(C)] = 0
    C[np.isinf(C)] = 0
    evalues, evectors = np.linalg.eig(C)
    mappedX = X @ np.real(evectors)[:, :no_dims]
    return mappedX



import numpy as np


def init_pca(X, no_dims, contri):
    """
    This function preprocesses data with excessive size and dimensions.

    Parameters are:

    'X'          - N by D matrix. Each row in X represents an observation.
    'no_dims'    - A positive integer specifying the number of dimension of the representation Y.
    'contri'     - Threshold of PCA variance contribution.

    """
    m = X.shape[1]
    X = X - np.mean(X, axis=0)
    # Compute covariance matrix C
    C = np.cov(X, rowvar=False)
    # Compute covariance matrix C
    C[np.isnan(C)] = 0
    C[np.isinf(C)] = 0
    lamda, M = np.linalg.eig(C)
    lamda = np.real(lamda)
    # Obtain the best PCA dimension
    if m < 2001:
        ind = np.where(np.cumsum(lamda) / sum(lamda) > contri)
    else:
        ind = np.where(np.cumsum(lamda) / sum(lamda[:2000]) > contri)
    bestDim = max(no_dims + 1, int(ind[0][0]))
    # Apply mapping on the data
    mappedX = X @ np.real(M)[:, :bestDim]

    return mappedX



from scipy.spatial.distance import cdist
import numpy as np


def mds(X, no_dims):
    """
      This function performs MDS embedding.

      Parameters are:

      'X'          - N by D matrix. Each row in X represents an observation.
      'no_dims'    - A positive integer specifying the number of dimension of the representation Y.

    """
    n = X.shape[0]
    D = cdist(X, X) ** 2
    sumd = np.mean(D, axis=1)
    sumD = np.mean(sumd)
    B = np.zeros((n, n))
    for i in range(n):
        for j in range(i+1, n):
            B[i][j] = -0.5 * (D[i][j] - sumd[i] - sumd[j] + sumD)
            B[j][i] = B[i][j]
    value, U = np.linalg.eig(B)
    embedX = U[:, :no_dims] @ np.diag(np.sqrt(np.abs(value[:no_dims])))
    return embedX



from sklearn.neighbors import NearestNeighbors
from scipy.spatial.distance import cdist
import numpy as np


def opt_scale(X, Y, k_num):
    """
      This function computes the optimal scales of landmarks.

      Parameters are:

      'X'       - N by D matrix. Each row in X represents high-dimensional features of a landmark.
      'Y'       - N by d matrix. Each row in Y represents low-dimensional embedding of a landmark.
      'k_num'   - A non-negative integer specifying the number of KNN.

    """
    n = X.shape[0]
    get_knn = NearestNeighbors(n_neighbors=k_num + 1).fit(Y).kneighbors(Y, return_distance=False)
    scale = np.zeros((n, 1))
    for i in range(n):
        XDis = cdist(X[get_knn[i]], X[get_knn[i]])
        YDis = cdist(Y[get_knn[i]], Y[get_knn[i]])
        scale[i] = np.sum(XDis * YDis) / max(np.sum(XDis ** 2), np.finfo(float).tiny)

    return scale



from sklearn.neighbors import NearestNeighbors
from scipy.sparse import diags
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import eigs
from scipy.spatial.distance import cdist
from .init_pca import init_pca
from .pca import pca
from .mds import mds
import numpy as np


def learning_s(X_samp, k1, get_knn, rnn, id_samp, no_dims, initialize, agg_coef, T_epoch):
    """
    This function returns representation of the landmarks in the lower-dimensional space and the number of nearest
    neighbors of landmarks. It computes the gradient using the entire probability matrix P and Q.

    """
    # Obtain size and dimension of landmarks
    N, dim = X_samp.shape

    # Compute the number of nearest neighbors of landmarks adaptively
    if N < 9:
        k2 = N
    else:
        if N > 1000:
            k2 = int(np.ceil(np.log2(N)) + 18)
        elif N > 50:
            k2 = int(np.ceil(0.02 * N)) + 8
        else:
            k2 = 9

    # Compute high-dimensional probability matrix P
    if k1 > 0:
        # Compute SNN matrix of landmarks
        SNN = np.zeros((N, N))
        knn_rnn_mat = rnn[get_knn[id_samp]]
        for i in range(N):
            snn_id = np.isin(get_knn[id_samp], get_knn[id_samp[i]]).astype(int)
            nn_id = np.where(np.max(snn_id, axis=1) == 1)[0]
            SNN[i, nn_id] = np.sum(knn_rnn_mat[nn_id] * snn_id[nn_id], axis=1)
            SNN[i] = SNN[i] / max(np.max(SNN[i]), np.finfo(float).tiny)
        # Compute the modified distance matrix
        Dis = (1 - SNN) ** agg_coef * cdist(X_samp, X_samp)
        P = np.zeros((N, N))
        sort_dis = np.sort(Dis, axis=1)
        idx = np.argsort(Dis, axis=1)
        for i in range(N):
            P[i, idx[i, :k2]] = np.exp(
                -0.5 * np.square(sort_dis[i, :k2]) / np.maximum(np.square(np.mean(sort_dis[i, :k2])),
                                                                np.finfo(float).tiny))
    else:
        if N > 5000 and dim > 50:
            xx = init_pca(X_samp, no_dims, 0.8)
            samp_dis, samp_knn = NearestNeighbors(n_neighbors=k2).fit(xx).kneighbors(xx)
        else:
            samp_dis, samp_knn = NearestNeighbors(n_neighbors=k2).fit(X_samp).kneighbors(X_samp)
        mean_samp_dis_squared = np.square(np.mean(samp_dis, axis=1))
        Pval = np.exp(-0.5 * np.square(samp_dis) / np.maximum(mean_samp_dis_squared[:, np.newaxis], np.finfo(float).tiny))
        P = csr_matrix((Pval.flatten(), ([i for i in range(N) for _ in range(k2)], samp_knn.flatten())), shape=(N, N)).toarray()
    # Symmetrize matrix P
    P = (P + P.transpose()) / 2

    # Initialize embedding Y of landmarks
    if initialize == 'le':
        Dg = diags(np.array(np.sum(P, axis=0)))
        L = np.sqrt(Dg) @ (Dg - P) @ np.sqrt(Dg)
        eigenvalues, eigenvectors = eigs(L, k=no_dims + 1, which='SM')
        smallest_indices = np.argsort(np.abs(eigenvalues))
        Y = np.real(eigenvectors[:, smallest_indices[1:]])
        del Dg, L
    elif initialize == 'pca':
        Y = pca(X_samp, no_dims)
    elif initialize == 'mds':
        Y = mds(X_samp, no_dims)

    # Normalize matrix P
    P = P / (np.sum(P) - N)

    # Initialization
    max_alpha = 2.5 * N
    min_alpha = 2 * N
    warm_step = 10
    preGrad = np.zeros((N, no_dims))
    epoch = 1
    while epoch <= T_epoch:
        # Update learning rate
        if epoch <= warm_step:
            alpha = max_alpha
        else:
            alpha = min_alpha + 0.5 * (max_alpha - min_alpha) * (
                        1 + np.cos(np.pi * ((epoch - warm_step) / (T_epoch - warm_step))))
        # Update matrix Q
        D = cdist(Y, Y) ** 2
        Q1 = 1 / (1 + np.log(1 + D))
        QQ1 = 1 / (1 + D)
        Q = Q1 / (np.sum(Q1) - N)
        # Compute gradient
        ProMatY = 4 * (P - Q) * Q1 * QQ1
        grad = (np.diag(np.sum(ProMatY, axis=0)) - ProMatY) @ Y
        # Update embedding Y
        Y = Y - alpha * (grad + (epoch - 1) / (epoch + 2) * preGrad)
        preGrad = grad
        # Compute KLD cost
        epoch = epoch + 1

    print(str(epoch - 1) + ' epochs have been computed!')
    return Y, k2


import numpy as np


def clle(X_samp, Y_samp, X_i, N_dis):
    """
    Constrained Locally Linear Embedding (CLLE)
    This function returns representation of point X_i.

    Parameters are:

    'X_samp'      - High-dimensional features of KNN of point X_i. Each row denotes an observation.
    'Y_samp'      - Low-dimensional embeddings of KNN of point X_i.
    'X_i'         - Current non-landmark point.
    'N_dis'       - Distance between point X_i and its nearest neighbor in lower-dimensional space.

    """
    n = X_samp.shape[0]
    S = (X_samp - X_i) @ (X_samp - X_i).transpose()
    if np.abs(np.linalg.det(S)) <= np.finfo(float).eps:
        S = S + (0.1 ** 2 / n) * np.trace(S) * np.eye(n)
    W = (np.linalg.inv(S) @ np.ones((n, 1))) / (np.ones((1, n)) @ np.linalg.inv(S) @ np.ones((n, 1)))
    Y_0 = W.transpose() @ Y_samp
    dd = np.sqrt((Y_samp[0] - Y_0) @ (Y_samp[0] - Y_0).transpose())
    if dd != 0:
        Y_i = Y_samp[0] + N_dis * (Y_0 - Y_samp[0]) / dd
    else:
        Y_i = Y_samp[0]

    return Y_i